"""
Utilities for unittests
"""

# import multiprocessing
import threading
from smartrpyc import server


class ExampleRpcProcess(threading.Thread):
    """Process running a SmartRPyC server"""

    def __init__(self, methods, addresses=None, middleware=None):
        super(ExampleRpcProcess, self).__init__()
        self._methods = methods
        self._addresses = addresses
        self._middleware = middleware

    def run(self):
        self.rpc = server.Server(self._methods)
        if self._middleware is not None:
            self.rpc.middleware[:] = self._middleware[:]

        self.rpc.bind(self._addresses)
        self.rpc.run()


class TestingServer(object):
    """Context manager to provide a server"""

    def __init__(self, methods, addresses, middleware=None):
        self.methods = methods
        self.addresses = addresses
        self.middleware = middleware

    def __enter__(self):
        self.rpc_process = ExampleRpcProcess(
            self.methods, self.addresses, middleware=self.middleware)
        self.rpc_process.daemon = True
        self.rpc_process.start()
        return self.rpc_process

    def __exit__(self, exc_type, exc_val, exc_tb):
        # self.rpc_process.terminate()
        # self.rpc_process.join(timeout=3)
        ## the thread should terminate when going out of scope..
        pass



# from smartrpyc.tests.utils import unittest

import pytest
from smartrpyc.server import MethodsRegister


class TestMethodsRegister(object):

    def test_registration(self):
        register = MethodsRegister()

        func = lambda x: x
        register.register(func)
        register.register(name="test-x")(func)
        register.register(func, name="test-y")
        assert register.lookup("<lambda>") == func
        assert register.lookup("test-x") == func
        assert register.lookup("test-y") == func

    def test_no_callable(self):
        register = MethodsRegister()
        with pytest.raises(ValueError):
            register.register("not_callable")
            register.register("not_callable", "test")



"""
"Public API" client
"""

from smartrpyc.client import Client


class PublicClient(Client):
    """
    Client for the public API.
    Includes what needed to handle authentication tokens.
    """

    def __init__(self, *a, **kw):
        super(PublicClient, self).__init__(*a, **kw)
        self._token = None

    def _prepare_request(self, request):
        if self._token is not None:
            request['t'] = self._token
        return request

    def login(self, username, password):
        """Performs a login, storing token to be used in further requests"""
        self._token = self.authenticate(username, password)

    def set_token(self, token):
        """Directly set the token to be used for authentication"""
        self._token = token

    def logout(self):
        """Logs out, removing authorization token"""
        self._token = None



"""
Server middleware
"""
import logging
from smartrpyc.server import DirectResponse

__all__ = ['ServerMiddlewareBase', 'IntrospectionMiddleware']

logger = logging.getLogger(__name__)


class ServerMiddlewareBase(object):
    """
    Base class for the server middleware.

    This is mostly a reminder for methods signatures, but subclassing
    from this object is not mandatory for middleware classes, as long
    as they expose a ``pre()`` and/or ``post()`` methods.
    """

    def pre(self, request, method):
        """
        This method will be executed **before** the request is processed.

        :param request:
            a :py:class:`smartrpyc.server.Request` object
        :param method:
            the method that is going to be called to process the request
            and generate the response

        This method has different ways to hijack the request/response
        process:

        * Raise a :py:exc:`~smartrpyc.server.DirectResponse` exception,
          containing some object to be directly returned to the client

        * Raise a :py:exc:`~smartrpyc.server.SetMethod` exception,
          containing an alternate method to be called

        * Raise a generic exception that will terminate the request
          processing and will be returned to the client directly.
        """
        pass

    def post(self, request, method, response, exception):
        """
        This method will be executed **after** the request is processed
        and a response returned by the called method.

        :param request:
            a :py:class:`smartrpyc.server.Request` object
        :param method:
            the method that has been called to process the request
            and generate the response
        :param response:
            the response object returned from the method call (if any)
        :param exception:
            the exception raised by the method call (if any)
        """
        pass


class IntrospectionMiddleware(object):
    """
    Middleware adding the ``.dir()`` and ``.doc()`` methods to be used
    for introspecting the exposed methods.
    """

    def pre(self, request, method):
        ## Note: this is experimental, method names may change!
        if request.method == 'dir':
            logger.debug('Called .dir() method')
            raise DirectResponse(request.server.methods.list_methods())
        if request.method == 'doc':
            logger.debug('Called .doc() method for method {0}'
                         ''.format(request.args[0]))
            _method = request.server.methods.lookup(request.args[0])
            raise DirectResponse(_method.__doc__)



"""
Exceptions returned to the client from the RPC API.
"""


class RpcServiceException(Exception):
    pass


class LoginFailed(RpcServiceException):
    pass


class LoginRequired(RpcServiceException):
    pass


class NotAuthorized(RpcServiceException):
    pass



"""
Base objects for the RPC
"""

import logging

import zmq

from smartrpyc.utils import lazy_property
from smartrpyc.utils.serialization import MsgPackSerializer
from .register import MethodsRegister
from .exceptions import DirectResponse, SetMethod

__all__ = ['Server', 'Request']

logger = logging.getLogger(__name__)


class Request(object):
    """Wrapper for requests from the RPC"""

    server = None

    def __init__(self, raw):
        """
        :param raw: The (unpacked) content of the request
        """
        self.raw = raw

    @property
    def id(self):
        """Id of the request"""
        return self.raw["i"]

    @property
    def method(self):
        """Name of the method to be called"""
        return self.raw['m']

    @property
    def args(self):
        """Positional arguments for the called method"""
        return self.raw.get('a') or ()

    @property
    def kwargs(self):
        """Keyword arguments for the called method"""
        return self.raw.get('k') or {}


class Server(object):
    request_class = Request
    packer = MsgPackSerializer

    def __init__(self, methods=None):
        """
        Constructor for the RPC Server class

        :param methods:
            Methods to be exposed via RPC. Usually a ``MethodsRegister``
            instance, but this is not mandatory.
            It has to have a ``lookup(name)`` method, returning
            the specified method, or raising an exception if the method
            was not found.
        """
        self.methods = methods
        if self.methods is None:
            self.methods = MethodsRegister()

        self.middleware = []  # Middleware chain

    @lazy_property
    def socket(self):
        context = zmq.Context()
        return context.socket(zmq.REP)

    def bind(self, addresses):
        """
        Bind the server socket to an address (or a list of)

        **Beware!** Only ``.bind()`` or ``.connect()`` may be called on
        a given server instance!

        :params addresses:
            A (list of) address(es) to which to bind the server.
        """
        if not isinstance(addresses, (list, tuple)):
            addresses = [addresses]
        for addr in addresses:
            self.socket.bind(addr)

    def connect(self, addresses):
        """
        Connect the server socket to an address (or a list of)

        **Beware!** Only ``.bind()`` or ``.connect()`` may be called on
        a given server instance!

        :params addresses:
            A (list of) address(es) to which to connect the server.
        """
        if not isinstance(addresses, (list, tuple)):
            addresses = [addresses]
        for addr in addresses:
            self.socket.connect(addr)

    def run(self):
        """Start the server listening loop"""
        while True:
            self.run_once()

    def run_once(self):
        """Run once: process a request and send a response"""
        message_raw = self.socket.recv()
        request = self.request_class(self.packer.unpackb(message_raw))
        request.server = self
        response = self._process_request(request)
        self.socket.send(self.packer.packb(response))

    def _process_request(self, request):
        """Process a received request"""

        logger.debug("Processing request")

        exception = None
        method = None

        ## Lookup the requested method
        try:
            method = self.methods.lookup(request.method)
        except KeyError:
            msg = 'No such method: {0}'.format(request.method)
            logger.error(msg)
            exception = KeyError(msg)

        ## Execute all the PRE middleware on the request
        try:
            self._exec_pre_middleware(request, method)

        except DirectResponse, e:
            logger.debug("A PRE middleware requested to send a response now")
            return self._response_message(e.response)

        except SetMethod, e:
            logger.debug("A PRE middleware changed the method to be called")
            method = e.method
            exception = None  # Clear exceptions happened before..

        except Exception, e:
            logger.exception('Exception during execution of PRE middleware')
            return self._exception_message(e)

        ## If the method is still None, return the current exception
        if method is None:
            if exception is None:
                exception = KeyError("No such method")
            return self._exception_message(exception)

        ## Execute the method
        try:
            response = method(request, *request.args, **request.kwargs)
        except Exception, e:
            logger.exception('Exception during execution of request')
            response = None
            exception = e
        else:
            exception = None

        ## Execute all the POST middleware on request + response
        try:
            response = self._exec_post_middleware(
                request, method, response, exception)
        except DirectResponse, e:
            logger.info(
                "A POST middleware requested immediate returning "
                "of a response")
            return self._response_message(e.response)
        except Exception, e:
            logger.exception('Exception during execution of POST middleware')
            return self._exception_message(e)

        ## If there was an exception, return it as a special message
        if exception is not None:
            return self._exception_message(exception)

        ## Return the response message
        return self._response_message(response)

    def _exception_message(self, exception):
        return {
            'e': type(exception).__name__,
            'e_msg': str(exception),
        }

    def _response_message(self, response):
        return {'r': response}

    def _exec_pre_middleware(self, request, method):
        for mw in self.middleware:
            if hasattr(mw, 'pre'):
                mw.pre(request, method)

    def _exec_post_middleware(self, request, method, response, exception):
        for mw in reversed(self.middleware):
            if hasattr(mw, 'post'):
                retval = mw.post(request, method, response, exception)
                if retval is not None:
                    response = retval
        return response



import pytest
from smartrpyc.server import MethodsRegister
from smartrpyc.client import pirate
from smartrpyc.client import exceptions
from smartrpyc.utils import get_random_ipc_socket
from smartrpyc.tests import utils


class TestLazyClient(object):

    def get_methods(self):
        self.methods = MethodsRegister()

        @self.methods.register
        def hello(request):
            return "world"

        return self.methods

        # self.addr = get_random_ipc_socket()

    def test_server_active(self):
        addr = get_random_ipc_socket()

        with utils.TestingServer(self.get_methods(), addr) as proc:
            proc.join(1)

            # retries 3, timeout 3
            c = pirate.Lazy(address=addr)
            assert 'world' == c.hello()

    def test_server_unavailable(self):
        # retries 3, timeout 3
        addr = get_random_ipc_socket()
        c = pirate.Lazy(retries=1, timeout=0, address=addr)

        with pytest.raises(exceptions.ServerUnavailable):
            c.hello()



"""
Serialization utilities
"""
import json
import pickle

import msgpack
from msgpack.fallback import Unicode


__all__ = ['MsgPackSerializer', 'CustomMsgPackSerializer', 'JsonSerializer',
           'PickleSerializer']


class MsgPackSerializer(object):
    @staticmethod
    def packb(o):
        return msgpack.packb(o, encoding='utf-8')

    @staticmethod
    def unpackb(packed):
        return msgpack.unpackb(packed, encoding='utf-8')


class CustomMsgPackSerializer(object):
    """
    Custom msgpack serializer, allowing to differentiate between
    strings and blobs. It's way more inefficient thant the MsgPackSerializer,
    however, since it has to use the fallback Python-only version..
    """
    @staticmethod
    def packb(o):
        return msgpack.packb(_pack_strings(o))

    @staticmethod
    def unpackb(packed):
        return _unpack_strings(msgpack.unpackb(packed))


def _pack_strings(o):
    if isinstance(o, Unicode):
        return b"u" + o.encode('utf-8')
    elif isinstance(o, bytes):
        return b"b" + o
    elif isinstance(o, dict):
        return dict(
            (_pack_strings(k), _pack_strings(v))
            for k, v in o.iteritems())
    elif isinstance(o, (list, tuple)):
        return [_pack_strings(x) for x in o]
    else:
        return o


def _unpack_strings(o):
    if isinstance(o, bytes):
        if o[0] in (b'u', ord(b'u')):
            return o[1:].decode('utf-8')
        elif o[0] in (b'b', ord(b'b')):
            return o[1:]
        raise ValueError("Invalid string object prefix {0}!".format(o[0]))
    elif isinstance(o, dict):
        return dict(
            (_unpack_strings(k), _unpack_strings(v))
            for k, v in o.iteritems())
    elif isinstance(o, (list, tuple)):
        return [_unpack_strings(x) for x in o]
    else:
        return o


class JsonSerializer(object):
    """
    .. warning::
        We have a problem with blobs here, since all the strings
        are automatically converted to unicode...
    """
    @staticmethod
    def packb(o):
        return json.dumps(o)

    @staticmethod
    def unpackb(packed):
        return json.loads(packed)


class PickleSerializer(object):
    """
    Pickle-powered serializer

    .. warning::
        Never, ever, use this for untrusted data!! Big security risk!!
    """
    @staticmethod
    def packb(o):
        return pickle.dumps(o)

    @staticmethod
    def unpackb(packed):
        return pickle.loads(packed)



"""
Method Registers module

This module provides facilities for preparing a library of methods
to be exposed via the RPC.

The library creation can happen in several ways:

* Using a :py:class:`.MethodsRegister` decorator, to register a bunch
  of stand-alone functions
* Usinng a :py:class:`.MethodsObject` to expose all the methods in a
  given object (instance). This is especially useful if you need
  to expose methods from an object **instance**, ie. you need to share
  some state.
"""

__all__ = ['MethodsRegisterBase', 'MethodsRegister', 'MethodsObject']


class MethodsRegisterBase(object):
    """
    Base for the Method Register objects.

    This class provides an abstract of the methods that can (and should)
    be exposed by a methods register object.
    """

    def register(self, func=None, name=None):
        """Only used for registers supporting changes"""
        raise NotImplementedError

    def lookup(self, method):
        """
        The most important method, returning the appropriate
        method to be called.
        """
        raise NotImplementedError

    def list_methods(self):
        """List all the supported methods"""
        raise NotImplementedError


class MethodsRegister(MethodsRegisterBase):
    """
    Register for methods to be exposed via RPC.
    Mostly a wrapper around a dict.

    Usage::

        methods = MethodsRegister()

        @methods.register
        def my_method(request, arg1, arg2):
            pass

        @methods.register(name='method2')
        def my_other_method(request, arg1=123, arg2=None):
            pass

        methods.register(func=func, name='test')

        # then pass methods to the Server() constructor
    """
    def __init__(self):
        self._methods = {}

    def register(self, func=None, name=None):
        """
        Refer to class docstring

        :params func: Callable to register
        :params name: Basestring to use as reference
            name for `func`
        """

        # If name is None we'll get the name from the
        # function itself.
        name = name or callable(func) and func.__name__

        def decorator(func):
            self.store(name, func)
            return func

        # If func is not None, it means the method
        # has been used either as a simple decorator
        # or as a plain method. The function will be
        # registered by calling decorator, otherwise,
        # decorator will be returned.
        return func and decorator(func) or decorator

    def store(self, name, function):
        """
        Method used to finally register a
        function.

        :params name:
            Functions lookup name
        :params function:
            Callable function
        """
        if not isinstance(name, basestring) and \
                not callable(function):
            # If we get here, most likely, the user has passed
            # a non callable function. See unittest for a clearer
            # example.
            raise ValueError("Unsupported arguments to register: "
                             "{0} and {1}".format(name, function))
        self._methods[name] = function

    def lookup(self, method):
        return self._methods[method]

    def list_methods(self):
        return self._methods.keys()


class MethodsObject(MethodsRegisterBase):
    """
    Objects wrapper register.

    This is a "wrapper" register, that automatically exposes all the methods
    contained in the object passed to the constructor.
    """

    _object = None

    def lookup(self, method):
        return getattr(self._object, method)

    def list_methods(self):
        return [m for m in dir(self._object) if not m.startswith('_')]

    def set_object(self, obj):
        if self._object is not None:
            raise RuntimeError("You cannot set the object twice.")
        self._object = obj



"""
"Public API" server

.. note::
    This is just a stub, taken from another project that contains
    a fully-featured implementation of the thing.. Feel free to copy
    in your project or subclass this.
"""

from smartrpyc.utils import lazy_property
from smartrpyc.exceptions import LoginRequired, NotAuthorized
from smartrpyc.server import Request, Server, ServerMiddlewareBase


class PublicRequest(Request):
    """
    Request class from the Public RPC (supports authentication)
    """

    @lazy_property
    def token(self):
        return self.raw.get('t')

    @lazy_property
    def token_info(self):
        ## todo: return information about the authorization token
        ##       such as permissions and expiration date
        pass

    @property
    def user_id(self):
        return self.token_info['user_id']

    @lazy_property
    def user(self):
        if self.token is None:
            return None
        ## todo: return information on the user linked to the token

    def is_logged_in(self):
        return self.user is not None

    def check_permissions(self, permissions):
        ## todo: Make sure the token id valid and grants the selected perms
        pass


class AuthCheckMiddleware(ServerMiddlewareBase):
    def pre(self, request, method):
        if hasattr(method, 'login_required') and method.login_required:
            if not request.is_logged_in():
                raise LoginRequired("You must be logged in.")
        if hasattr(method, 'require_permissions'):
            try:
                request.check_permissions(method.require_permissions)
            except:
                raise NotAuthorized("You miss some required permission(s)")


class PublicServer(Server):
    request_class = PublicRequest

    def __init__(self, *a, **kw):
        super(PublicServer, self).__init__(*a, **kw)
        self.middleware.append(AuthCheckMiddleware())


def login_required(fun):
    """
    Decorator for RPC methods -- requires login
    """
    fun.login_required = True
    return fun


def require_permissions(*permissions):
    """
    Decorator for RPC methods -- requires permissions
    """
    def decorator(fun):
        fun.require_permissions = permissions
        return fun
    return decorator



"""
Exceptions used by the Server.
"""

__all__ = ['DirectResponse', 'SetMethod']


class DirectResponse(Exception):
    """
    Exception used by middleware to directly send a response to the client.

    Request processing will be terminated immediately and the response
    sent to the client, as a normal response.
    """
    def __init__(self, response):
        self.response = response


class SetMethod(Exception):
    """
    Exception used by middleware to change the method to be executed.

    As a side effect, this clears all the previous "method not found"
    exceptions.
    """
    def __init__(self, method):
        self.method = method



# coding=utf-8
"""
Tests for serialization
"""


class TestSerialization(object):
    def _run_serialization_tests(self, packer, support_bytes=True):
        test_values = [
            ## These should present no significant problems!
            'Hello, world!',
            ['hello', 'world'],
            {'Hello': 'world'},
            {'Hello': 'world', 'Hello2': 'world2'},
            {'hello': ['w', 'o', 'r', 'l', 'd']},
            True, False, None,
            (2 ** 64) - 1,  # MAXINT (64bit) <-- Beware!
            10, 1.0, 0.2, -100000, -10.5,
            u"This is a unicode string",
            ['This is', u'a mixed', 'list'],
            u"Hélló wórld ¶¬¹²³¤€˛¡¿",
            {"this": u"is", "a": u"díctíöñáry"},
            {
                'ok, we can even do this': [
                    u"Hello, world!",
                    u"Hëllö, wørlḋ",
                    ['Hello', 'world'],
                ],
                u"And this??": {'yay': 'hello'},
                u"øf cøürsë, wé cán..": True,
                "ready?": [
                    10, 9, 8, 7, 6, 5, 5.5, 5.4, 5.3, 5.2, "Boom!", None,
                ],
            },
            u"\u1234\u5678 <-- this is... something!",
            u"What's this \x01 \x02 \x03 \x80\x81\x92\x93 crap??",
        ]

        ## These should work only with msgpack_custom and pickle
        extended_test_values = [
            "\x01\x02\x03\x04\x05\x06",
            b"And this is \x00\x01\x02\x03 a py3k \x80\x81\x82\x83 binary",
        ]

        if support_bytes:
            test_values += extended_test_values

        for value in test_values:
            _new_value = packer.unpackb(packer.packb(value))
            assert value == _new_value
            if support_bytes:
                assert type(value) == type(_new_value)

    def test_serialization_msgpack(self):
        from smartrpyc.utils.serialization import MsgPackSerializer
        self._run_serialization_tests(MsgPackSerializer, support_bytes=False)

    def test_serialization_msgpack_custom(self):
        from smartrpyc.utils.serialization import CustomMsgPackSerializer
        self._run_serialization_tests(CustomMsgPackSerializer)

    def test_serialization_json(self):
        from smartrpyc.utils.serialization import JsonSerializer
        self._run_serialization_tests(JsonSerializer, support_bytes=False)

    def test_serialization_pickle(self):
        from smartrpyc.utils.serialization import PickleSerializer
        self._run_serialization_tests(PickleSerializer)



# coding=utf-8
"""
Tests for SmartRPC
"""

import random

import zmq
import msgpack
import pytest

from smartrpyc.client import Client, RemoteException
from smartrpyc.server import MethodsRegister
from smartrpyc.utils import get_random_ipc_socket
from smartrpyc.tests import utils


# noinspection PyUnusedLocal
class TestSmartrpcFunctionality(object):
    def get_methods(self):

        methods = MethodsRegister()

        @methods.register
        def method1(request):
            return u"Hello, world!"

        @methods.register
        def method2(request, name):
            return u"Hello, {0}!".format(name)

        @methods.register
        def method3(request, greeting=u'Hello', name=u'world'):
            if not isinstance(greeting, basestring):
                raise ValueError("Greeting must be a string")
            if not isinstance(name, basestring):
                raise ValueError("Name must be a string")
            return u"{0}, {1}!".format(greeting, name)

        @methods.register
        def get_list(request):
            return ['this', 'is', 'a', 'list']

        @methods.register
        def get_dict(request):
            return {'this': 'is', 'a': 'dict'}

        @methods.register
        def raise_value_error(request, *a, **kw):
            raise ValueError

        return methods

    def test_simple_call(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(addr)
            socket.send(msgpack.packb({
                'm': 'method1',
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'Hello, world!'}

    def test_concurrent_calls(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):

            context = zmq.Context()
            sockets = {}
            socket_ids = range(10)

            random.shuffle(socket_ids)
            for i in socket_ids:
                sockets[i] = context.socket(zmq.REQ)
                sockets[i].connect(addr)

            random.shuffle(socket_ids)
            for i in socket_ids:
                sockets[i].send(msgpack.packb({
                    'm': 'method2',
                    'a': ['socket {0}'.format(i)],
                }, encoding='utf-8'))

            random.shuffle(socket_ids)
            for i in socket_ids:
                response = msgpack.unpackb(sockets[i].recv(), encoding='utf-8')
                assert response == {'r': 'Hello, socket {0}!'.format(i)}

    def test_call_with_args(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(addr)
            socket.send(msgpack.packb({
                'm': 'method2',
                'a': ['WORLD'],
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'Hello, WORLD!'}

    def test_call_with_args_utf8(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):

            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(addr)

            socket.send(msgpack.packb({
                'm': 'method2',
                'a': [u'WORLD'],
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': u'Hello, WORLD!'}

            socket.send(msgpack.packb({
                'm': 'method2',
                'a': [u'Ẅøřłđ'],
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': u'Hello, Ẅøřłđ!'}

    def test_call_with_kwargs(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(addr)
            socket.send(msgpack.packb({
                'm': 'method2',
                'k': {'name': 'WORLD'},
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'Hello, WORLD!'}

    def test_call_with_multi_args(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect(addr)

            socket.send(msgpack.packb({
                'm': 'method3',
                'k': {'name': 'WORLD'},
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'Hello, WORLD!'}

            socket.send(msgpack.packb({
                'm': 'method3',
                'k': {'greeting': 'HOWDY'},
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'HOWDY, world!'}

            socket.send(msgpack.packb({
                'm': 'method3',
                'k': {'greeting': 'HOWDY', 'name': 'MAN'},
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'HOWDY, MAN!'}

            socket.send(msgpack.packb({
                'm': 'method3',
                'a': ('hey', 'man'),
            }, encoding='utf-8'))
            response = msgpack.unpackb(socket.recv(), encoding='utf-8')
            assert response == {'r': 'hey, man!'}

    def test_with_client(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(methods=self.get_methods(), addresses=addr):
            client = Client(addr)
            assert client.method1() == 'Hello, world!'

            with pytest.raises(RemoteException):
                client.method1('this', 'does not', 'accept arguments!')

            assert client.method2(u'WORLD') == u'Hello, WORLD!'
            assert client.method2(name=u'WORLD') == u'Hello, WORLD!'
            assert client.method2(name=u'Ẅøřłđ') == u'Hello, Ẅøřłđ!'

            assert client.method3('Hi', 'man') == 'Hi, man!'
            assert client.method3(greeting='Hi', name='man') == 'Hi, man!'
            assert client.method3(greeting='Hi') == 'Hi, world!'

            with pytest.raises(RemoteException):
                client.method3('HELLO', greeting='Hi')

            assert client.get_list() == ['this', 'is', 'a', 'list']
            assert client.get_dict() == {'this': 'is', 'a': 'dict'}

            with pytest.raises(RemoteException):
                client.raise_value_error()

            ## Ok, the method doesn't exist, no big deal..
            with pytest.raises(RemoteException):
                client.no_such_method()

            client.method1()  # The server must still be alive here..



import zmq

from smartrpyc import utils
from smartrpyc.client import base, exceptions


class Lazy(base.Client):
    def __init__(self, retries=3, timeout=1000, **kwargs):
        self._retries = retries
        self._timeout = timeout
        super(Lazy, self).__init__(**kwargs)

    @utils.lazy_property
    def _poller(self):
        poller = zmq.Poller()
        poller.register(self._socket, zmq.POLLIN)
        return poller

    def _get_response(self, request, retries=None):
        if not retries is None and retries <= 0:
            self._socket.setsockopt(zmq.LINGER, 0)
            self._socket.close()
            self._poller.unregister(self._socket)
            raise exceptions.ServerUnavailable("")

        socks = dict(self._poller.poll(self._timeout))
        if socks.get(self._socket) == zmq.POLLIN:
            return super(Lazy, self)._get_response(request)

        retries = (retries or self._retries) - 1
        return self._get_response(request, retries=retries)



"""
Utilities
"""

import os
import tempfile


def lazy_property(fn):
    attr_name = '_lazy_' + fn.__name__

    def getter(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    def setter(self, value):
        setattr(self, attr_name, value)

    def deleter(self):
        delattr(self, attr_name)

    return property(fget=getter, fset=setter, fdel=deleter, doc=fn.__doc__)


def get_random_ipc_socket():
    with tempfile.NamedTemporaryFile(delete=False, suffix='.sock') as s:
        os.unlink(s.name)
    return 'ipc://{0}'.format(s.name)



## SmartRPyC setup.py

import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

## Import version without tainting sys.modules
version = __import__('smartrpyc').__version__
if 'smartrpyc' in sys.modules:
    del sys.modules['smartrpyc']

install_requires = [
    'distribute',
    'pyzmq',
    'msgpack-python',
]

extra = {}
if sys.version_info >= (3,):
    extra['use_2to3'] = True
    #extra['convert_2to3_doctests'] = ['src/your/module/README.txt']
    #extra['use_2to3_fixers'] = ['your.fixers']


class PyTest(TestCommand):
    test_package_name = 'smartrpyc'

    def finalize_options(self):
        TestCommand.finalize_options(self)
        _test_args = [
            '--verbose',
            '--ignore=build',
            '--pep8',
        ]
        extra_args = os.environ.get('PYTEST_EXTRA_ARGS')
        if extra_args is not None:
            _test_args.extend(extra_args.split())
        self.test_args = _test_args
        self.test_suite = True

    def run_tests(self):
        import pytest
        from pkg_resources import normalize_path, _namespace_packages

        # Purge modules under test from sys.modules. The test loader will
        # re-import them from the build location. Required when 2to3 is used
        # with namespace packages.
        if sys.version_info >= (3,) and \
                getattr(self.distribution, 'use_2to3', False):
            #module = self.test_args[-1].split('.')[0]
            module = self.test_package_name
            if module in _namespace_packages:
                del_modules = []
                if module in sys.modules:
                    del_modules.append(module)
                module += '.'
                for name in sys.modules:
                    if name.startswith(module):
                        del_modules.append(name)
                map(sys.modules.__delitem__, del_modules)

            ## Run on the build directory for 2to3-built code
            ## This will prevent the old 2.x code from being found
            ## by py.test discovery mechanism, that apparently
            ## ignores sys.path..
            ei_cmd = self.get_finalized_command("egg_info")

            ## Replace the module name with normalized path
            #self.test_args[-1] = normalize_path(ei_cmd.egg_base)
            self.test_args.append(normalize_path(ei_cmd.egg_base))

        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name='SmartRPyC',
    version=version,
    packages=find_packages(),
    url='http://xk0.github.io/SmartRPyC',
    license='Apache License, Version 2.0, January 2004',
    author='Samuele Santi - Flavio Percoco',
    author_email='samuele@samuelesanti.com - flaper87@flaper87.org',
    description='SmartRPyC is a ZeroMQ-based RPC library for Python',
    long_description='SmartRPyC is a ZeroMQ-based RPC library for Python',
    install_requires=install_requires,
    tests_require=['pytest'],
    test_suite='smartrpyc.tests',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Networking",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    package_data={'': ['README.rst', 'LICENSE']},
    cmdclass={'test': PyTest},
    **extra)



"""
SmartRPC client
"""

import uuid
import zmq

from smartrpyc.utils import lazy_property
from smartrpyc.utils.serialization import MsgPackSerializer

__all__ = ['Client', 'IntrospectableClient', 'RemoteException',
           'ClientMiddlewareBase']


class RemoteException(Exception):
    """
    Wrapper around a remote exception.
    Ideally, this should fake to be an instance of the remote
    exception too, but that's not the case yet.. :(

    Eg:

    >>> exc = RemoteException('ValueError', 'Invalid value')
    >>> isinstance(exc, RemoteException)
    True
    >>> isinstance(exc, ValueError)
    True
    """

    def __init__(self, exc, msg):
        self.message = msg
        self.original_exc = exc

    def __str__(self):
        return "{0}: {1}".format(self.original_exc, self.message)

    def __repr__(self):
        return "<RemoteException: {0!r}>".format(str(self))


class Client(object):
    packer = MsgPackSerializer

    def __init__(self, address=None):
        self._address = address
        self.middleware = []  # Middleware chain

    @lazy_property
    def _socket(self):
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        # if self._address is not None:
        socket.connect(self._address)
        return socket

    def _do_request(self, method, args, kwargs):
        request = self._prepare_request({
            'i': str(uuid.uuid4()),
            'm': method,
            'a': args,
            'k': kwargs,
        })

        request = self._exec_pre_middleware(request)

        packed = self.packer.packb(request)
        self._socket.send(packed)
        return request

    def _get_response(self, request):
        response = self.packer.unpackb(self._socket.recv())

        response = self._exec_post_middleware(request, response)

        if 'e' in response:
            raise RemoteException(response['e'], response.get('e_msg'))
        return response['r']

    def _prepare_request(self, request):
        return request

    def __getattr__(self, item):
        if item.startswith('_'):
            raise AttributeError(item)

        server = self

        class method_proxy(object):
            def __call__(self, *a, **kw):
                request = server._do_request(item, a, kw)
                return server._get_response(request)

            @lazy_property
            def __doc__(self):
                request = server._do_request('doc', (item,), {})
                return server._get_response(request)

        # def method_proxy(*a, **kw):
        #     request = self._do_request(item, a, kw)
        #     return self._get_response(request)

        return method_proxy()

    def _exec_pre_middleware(self, request):
        for mw in self.middleware:
            if hasattr(mw, 'pre'):
                retval = mw.pre(request)
                if retval is not None:
                    request = retval
        return request

    def _exec_post_middleware(self, request, response):
        for mw in reversed(self.middleware):
            if hasattr(mw, 'post'):
                retval = mw.post(request, response)
                if retval is not None:
                    response = retval
        return response


class IntrospectableClient(Client):
    def __dir__(self):
        methods = dir(super(IntrospectableClient, self))
        methods += self.dir()
        return methods


class ClientMiddlewareBase(object):
    """
    Base class for client middleware.
    (mostly a reminder for methods signatures, subclassing from
    this is not mandatory).
    """

    def pre(self, request):
        pass

    def post(self, request, response):
        pass



# coding=utf-8
"""
Tests for SmartRPyC introspection
"""

from smartrpyc.client import IntrospectableClient
from smartrpyc.server import MethodsRegister, IntrospectionMiddleware
from smartrpyc.tests import utils
from smartrpyc.utils import get_random_ipc_socket


class TestIntrospection(object):

    def get_methods(self):
        methods = MethodsRegister()

        def method1(request):
            """Returns the string "Hello, world!"."""
            return u"Hello, world!"

        def method2(request, name):
            """Docstring of method2()"""
            return u"Hello, {0}!".format(name)

        def method3(request, greeting=u'Hello', name=u'world'):
            """Docstring of method3()"""
            if not isinstance(greeting, basestring):
                raise ValueError("Greeting must be a string")
            if not isinstance(name, basestring):
                raise ValueError("Name must be a string")
            return u"{0}, {1}!".format(greeting, name)

        methods.register(method1)
        methods.register(method2)
        methods.register(method3)
        return methods

    def test_list_methods(self):
        addr = get_random_ipc_socket()
        with utils.TestingServer(
                self.get_methods(), addr,
                middleware=[IntrospectionMiddleware()]):
            client = IntrospectableClient(addr)
            exposed_methods = dir(client)
            for meth in ('method1', 'method2', 'method3'):
                assert meth in exposed_methods

            assert client.method1.__doc__ == \
                'Returns the string "Hello, world!".'
            assert client.method2.__doc__ == 'Docstring of method2()'
            assert client.method3.__doc__ == 'Docstring of method3()'
